#!/usr/bin/python
#
# LDoms
#
# Copyright (c) 2005, 2016, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Script to get PCI properties from MD (either from a file or live from
/dev/mdesc on a system)

"""

import getopt
import struct
import sys

HEADER_SIZE = 16
ELEMENT_SIZE = 16

FMT_HEADER = "!HH I I I"
FMT_ELEMENT = "!c B H I Q"


def print_properties(request, props):
    """
    Given a list of properties, print out the name/value pairs.

    """

    if props is None:
        return

    match = False
    for (p, v) in request["conditions"].items():
        match = (str(props.get(p)) == v)
        if not match:
            break

    if not match:
        return

    if request["properties"]:
        pnames = request["properties"]
    else:
        pnames = props.keys()

    for p in pnames:
        if request["name"]:
            fmt = "%(name)s=%(value)s"
        else:
            fmt = "%(value)s"

        v = props.get(p, "")
        if type(v) is list:
            v = " ".join(v)
        else:
            v = str(v)

        print fmt % {"name": p, "value": v}


def process(mdfile, request):
    """
    Process the given MD file

    """

    f = open(mdfile, "r")
    md = f.read()
    f.close()

    header = struct.unpack_from(FMT_HEADER, md)
    (_, _, node_bsz, name_bsz, _) = header

    offset_node = HEADER_SIZE
    offset_name = offset_node + node_bsz
    offset_data = offset_name + name_bsz

    props = None
    offset = offset_node
    tag = ""
    while tag != "\0" and offset < offset_name:

        element = struct.unpack_from(FMT_ELEMENT, md, offset)
        (tag, nlen, _, noff, value) = element
        name = struct.unpack_from("%ds" % nlen, md, offset_name + noff)[0]

        if tag == "\0" or tag == "E":
            print_properties(request, props)
            props = None
            offset += ELEMENT_SIZE
            continue

        if tag == "N":
            if props is not None:
                raise Exception("Unexpected node: %d %s" % (offset, element))

            if name != request["node"]:
                offset = offset_node + value * ELEMENT_SIZE
            else:
                props = {}
                offset += ELEMENT_SIZE
            continue

        if props is None:
            raise Exception("Unexpected element: %d %s" % (offset, element))

        if tag == "v":
            props[name] = value
        elif tag == "s":
            soff = value & 0xffffffff
            slen = value >> 32
            sval = struct.unpack_from("%ds" %
                (slen - 1), md, offset_data + soff)[0]
            props[name] = sval
        elif tag == "d":
            doff = value & 0xffffffff
            dlen = value >> 32
            dval = struct.unpack_from("%ds" %
                (dlen - 1), md, offset_data + doff)[0]
            props[name] = dval.split("\0")

        offset += ELEMENT_SIZE


def main():
    """
    Main entry point.

    sample usage:
        mdprop.py -v virtual-device name=virtual-network-switch,cfg-handle=0

    """

    mdfile = "/dev/mdesc"
    noname = False

    opts, args = getopt.getopt(sys.argv[1:], "f:v")

    for o, a in opts:
        if o == "-f":
            mdfile = a
        elif o == "-v":
            noname = True

    request = {}
    request["name"] = not noname
    request["node"] = args[0]
    request["conditions"] = dict(map(lambda x: x.split("="),
                                     args[1].split(",")))
    if (len(args) >= 3):
        request["properties"] = args[2].split(",")
    else:
        request["properties"] = []

    process(mdfile, request)

main()
