#!/bin/bash
#
# LDoms
#
# Copyright (c) 2013, 2016, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

LOG=/tmp/vswlock.log.$$
MDPROP=/usr/lib/ldoms/mdprop.py

export PATH=/sbin:/bin:/usr/sbin:/usr/bin

#
# Get the IP configuration of an interface
#
ip_get()
{
    local interface=$1
    local dhcp address broadcast netmask

    dhcp=`pgrep -f "dhclient.*$interface"`

    set -- $(ifconfig $interface | \
	sed -n  's/.*inet addr:\(.*\)  Bcast:\(.*\)  Mask:\(.*\)/\1 \2 \3/p')

    address=${1:-"0.0.0.0"}
    broadcast=${2:-"0.0.0.0"}
    netmask=${3:-"0.0.0.0"}

    if [[ -n "$dhcp" ]]; then
	echo "dhcp"
    elif [[ "$address" = "0.0.0.0" ]]; then
	echo "none"
    else
	echo "static"
    fi

    echo $address
    echo $broadcast
    echo $netmask

    # get routes (destination, gateway, mask)
    route -n | awk -v itf=$interface '$4=="UG" && $8==itf {print $1,$2,$3}'
}

ip_reset()
{
    local interface=$1
    local state=$2
    local dhcp

    echo "resetting IP on $interface"

    dhcp=`pgrep -f "dhclient.*$interface"`
    if [[ -n "$dhcp" ]]; then
	kill $dhcp
    fi

    ifconfig $interface 0.0.0.0
    ifconfig $interface $state
}

#
# Put an IP configuration on an interface.
#
ip_set()
{
    local interface=$1
    local config=$2
    local address=$3
    local broadcast=$4
    local netmask=$5
    local destination gateway mask

    shift 5

    ip_reset $interface up

    if [[ "$config" != "dhcp" && "$config" != "static" ]]; then
	echo "configuring $interface with no IP"
	return
    fi

    if [[ -n "$address" ]]; then
	echo "configuring $interface with ${address}/${netmask}"
	ifconfig $interface $address netmask $netmask broadcast $broadcast
    fi

    # add routes

    while true; do
	if [[ -z "$1" ]]; then
	    break
	fi
	destination=$1
	gateway=$2
	mask=$3
	shift 3
	echo "adding route $destination/$mask $gateway"
	route add -net $destination netmask $mask gw $gateway
    done

    if [[ "$config" = "dhcp" ]]; then
	echo "configuring $interface with DHCP"
	dhclient -nw $interface 200>&-
    else
	echo "configuring $interface with static configuration"
	ifconfig $interface up
    fi
}

bridge_exists()
{
    local bridge=$1
    local br

    br=$(brctl show | awk -v br=$bridge '$1==br {print}')
    [[ -n "$br" ]]

    #ovs-vsctl br-exists $bridge
    #[[ $? -eq 2 ]]
}

bridge_create()
{
    local bridge=$1

    echo "creating bridge $vsw"
    brctl addbr $bridge
    #ovs-vsctl add-br $bridge
}

bridge_destroy()
{
    local bridge=$1

    echo "destroying bridge $vsw"
    ip_reset $bridge down
    brctl delbr $bridge
    #ovs-vsctl del-br $bridge
}

bridge_list_ports()
{
    local bridge=$1

    brctl show $bridge | awk -v br=$bridge \
	'NR==1 {next} $1==br {if ($4!="") print $4;next} {print $1}'
    #ovs-vsctl list-ports $bridge
}

bridge_add_port()
{
    local bridge=$1
    local port=$2

    ip_reset $port up
    echo "adding port $port to $bridge"
    brctl addif $bridge $port
    #ovs-vsctl add-port $bridge $port
}

bridge_remove_port()
{
    local bridge=$1
    local port=$2

    echo "removing port $port from $bridge"
    brctl delif $bridge $port
    #ovs-vsctl del-port $bridge $port
    ip_reset $port down
}


#
# Convert a macaddr from a decimal number (returned by MDPROP)
# to standard form (i.e. AA:BB:CC:DD:EE:FF) which used
# by the ip command.
#
mac_convert()
{
    local mac=$1
    local str=""

    #
    # for first five octets, add any required leading zeros, use upper case
    # and add trailing ':'
    #
    for (( CNTR=48; CNTR>0; CNTR-=8 )); do
	str+=$(printf "%.2X:" $(( ($mac >> $CNTR) & 0xff )) )
    done

    # print final octet (with no trailing ':')
    str+=$(printf "%.2X" $(($mac & 0xff)) )

    echo -e $str
}

vsw_add_port()
{
    local handle=$1
    local vsw="vsw$handle"
    local interface=$2
    local pdev ipconf mac

    bridge_exists $vsw
    if [[ $? -ne 0  ]]; then
	bridge_create $vsw

	# configure physical device if there's one.
	pdev=`$MDPROP -v virtual-device \
	    name=virtual-network-switch,cfg-handle=$handle \
	    vsw-phys-dev`
	if [[ -n "$pdev" ]]; then
	    ipconf=$(ip_get $pdev)
	    bridge_add_port $vsw $pdev
	    ip_set $vsw $ipconf
	else
	    #
	    # If this is a routed vsw (i.e. no physical net-dev specified)
	    # configure the MD assigned macaddr if there's one.
	    #
	    mac=`$MDPROP -v virtual-device \
		name=virtual-network-switch,cfg-handle=$handle \
		local-mac-address`
	    if [[ -n "$mac" ]]; then
		# convert to macaddr to standard form
		mac=$(mac_convert $mac)
		ip link set dev $vsw address $mac
	    fi
	fi
    fi

    bridge_add_port $vsw $interface
}

vsw_remove_port()
{
    local handle=$1
    local vsw="vsw$handle"
    local interface=$2
    local port ipconf

    bridge_remove_port $vsw $interface

    #
    # Remove the vsw is there are no more ports. If there's only one
    # port left, and it is not a vif then that's the vsw net-dev and
    # we also remove it.
    #
    set -- $(bridge_list_ports $vsw)

    if [[ $# -gt 1 ]]; then
	return
    fi

    port=$1

    if [[ $# -eq 1 && "$port" != vif* ]]; then
	ipconf=$(ip_get $vsw)
	bridge_remove_port $vsw $port
	ip_reset $vsw up
	ip_set $port $ipconf
    fi

    bridge_destroy $vsw
}

runlock()
{
    (
	flock -x 200
	$*
    ) 200>/var/lock/vswcfg.lock
}

usage()
{
    echo "Usage: vswcfg.sh [-u]"
    echo "       vswcfg.sh [-d] [-l] <action> <devpath> <interface>"
    echo
    exit 2
}

main()
{
    local debug=0
    local udev=0
    local log=0
    local action devpath interface vswhdl

    while getopts "dlu" opt; do
	case "${opt}" in
	    d)
		debug=1
		;;
	    l)
		log=1
		;;
	    u)
		udev=1
		;;
	    *)
		usage
		;;
	esac
    done

    #
    # If the script is invoked from udev, then we need to daemonize
    # otherwise we can deadlock when using bridge commands because
    # they can modify the device tree.
    #
    if [[ $udev -ne 0 ]]; then
	nohup $0 -l $ACTION $DEVPATH $INTERFACE &
	exit 0
    fi

    if [[ $log -ne 0 ]]; then
	echo "logging to $LOG"
	exec 1>$LOG 2>&1
    fi

    echo "args: $*"
    echo "date: $(date)"

    if [[ $debug -ne 0 ]]; then
	env
	set -x
    fi

    shift $((OPTIND - 1))

    action=$1
    devpath=$2
    interface=$3

    if [[ -z "$action" || -z "$devpath" || -z "$interface" ]]; then
	usage
    fi

    vswhdl=$(echo $devpath | \
	sed 's!/devices/channel-devices/vsw-port-\([0-9][0-9]*\)-.*!\1!')

    case $action in
	add)
	    runlock vsw_add_port $vswhdl $interface
	    ;;

	remove)
	    runlock vsw_remove_port $vswhdl $interface
	    ;;
    esac
}

main $*
