#!/usr/bin/python
#
# LDoms
#
# Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Decode a Machine Description (MD)

"""

from ldoms.MD import MDParser, MD_PARSE_NEXT_ELEMENT, MD_PARSE_NEXT_NODE


class MDDecoder(MDParser):
    """
    Class that implements the decoding of an MD.

    """

    def __init__(self):
        """
        Initialize the attributes in the MDDecoder class

        """
        MDParser.__init__(self)
        self.node_map = None
        self.node_name = None
        self.node_index = None
        self.node_processor = None
        self.node_processor_data = None
        self.node_decoder_factory = None
        self.node_decoder_data = None
        self.node_decoder = None

    def node_begin(self, node_name, node_index):
        """
        Begin decoding a node.

        """
        # get a node decoder from the node decoder factory.
        node_decoder = self.node_decoder_factory(self.node_decoder_data,
                                                 node_name, node_index)

        self.node_decoder = node_decoder

        if node_decoder is not None:
            node_decoder.begin()
            # keep track of node being decoded
            self.node_name = node_name
            self.node_index = node_index
            next_step = MD_PARSE_NEXT_ELEMENT
        else:
            # skip to next node
            next_step = MD_PARSE_NEXT_NODE

        return next_step

    def node_end(self):
        """
        End decoding a node.

        """
        node_object = self.node_decoder.end()

        if self.node_processor:
            self.node_processor(self.node_processor_data,
                                self.node_name, node_object)

        self.node_map[self.node_index] = self.node_decoder
        self.node_name = None
        self.node_index = None
        self.node_decoder = None

        return MD_PARSE_NEXT_ELEMENT

    def prop(self, prop_name, prop_value):
        """
        Decode a value, string or data property.

        """
        self.node_decoder.prop(prop_name, prop_value)
        return MD_PARSE_NEXT_ELEMENT

    def prop_arc(self, arc_name, arc_ref):
        """
        Decode an arc property.

        """
        self.node_decoder.arc(arc_name, arc_ref)
        return MD_PARSE_NEXT_ELEMENT

    def node_resolve(self, node_index):
        """
        Resolve a node reference.

        """
        return self.node_map.get(node_index)

    def finalize(self):
        """
        Finalize the decoding of a MD.

        """
        for node_decoder in self.node_map.values():
            node_decoder.finalize(self)

    def decode(self, mdesc,
               node_decoder_factory, node_decoder_data,
               node_processor, node_processor_data):
        """
        Decode a MD.

        """
        self.node_decoder = None
        self.node_decoder_factory = node_decoder_factory
        self.node_decoder_data = node_decoder_data
        self.node_processor = node_processor
        self.node_processor_data = node_processor_data

        self.node_map = {}
        self.node_name = None
        self.node_index = None

        mdesc.md_parse(self)

        self.finalize()
