#!/usr/bin/python
#
# LDoms
#
# Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Utility functions to access sysfs file.

"""

import glob
import os

_SYSFS_PCI_DEVICES = "/sys/bus/pci/devices"
_SYSFS_PCIDEV_PATTERN = _SYSFS_PCI_DEVICES + "/*"
_SYSFS_PCIDEV_OBPPATH_PATTERN = _SYSFS_PCIDEV_PATTERN + "/obppath"
_SYSFS_PCIDEV_SRIOV_NUMVFS_PATTERN = _SYSFS_PCIDEV_PATTERN + "/sriov_numvfs"

#
# Map MD paths to sysfs paths
#
_SYSFS_MD_PATHMAP = None


def _sysfs_init():
    """
    Initialization.

    """
    # pylint: disable = W0603
    global _SYSFS_MD_PATHMAP

    if _SYSFS_MD_PATHMAP is not None:
        return

    _SYSFS_MD_PATHMAP = {}
    for obpfile in glob.glob(_SYSFS_PCIDEV_OBPPATH_PATTERN):
        obppath = sysfs_read(obpfile)
        pcidev = os.path.dirname(obpfile)
        _SYSFS_MD_PATHMAP[obppath] = pcidev


def _sysfs_filearg(args, nparams=0):
    """
    Process arguments of sysfs_read/write (sysfs_rw) functions.
    A sysfs_rw function can be called either with a full pathname
    to a file, for example:

      sysfs_rw(full_pathname, parameters...)

    Or with a sysfs directory name and a file name, for example:

      sysfs_rw(dir_name, file_name, parameters...)

    A sysfs_rw(*args) will call _sysfs_filearg(args, nparams) to
    figure out the full pathname of the targeted sysfs file. nparams
    indicate the number of other parameters expected by the function.

    """
    args = list(args)
    if len(args) == nparams + 1:
        filename = args[0]
        del args[0]
    elif len(args) == nparams + 2:
        filename = os.path.join(args[0], args[1])
        del args[0:2]
    else:
        raise Exception("invalid number of parameters")

    if nparams == 0:
        return filename
    else:
        return [filename] + args


def sysfs_read(*args):
    """
    Read the content of a sysfs file. If the file doesn't exist,
    this returns None.

    """
    filename = _sysfs_filearg(args)
    if not os.path.exists(filename):
        return None
    fd = open(filename)
    data = fd.read().strip()
    fd.close()
    return data


def sysfs_write(*args):
    """
    Write data to a sysfs file.

    """
    (filename, data) = _sysfs_filearg(args, 1)
    if not os.path.exists(filename):
        raise Exception("unknown file %s" % filename)
    fd = open(filename, "w")
    fd.write(str(data))
    fd.close()


def sysfs_readlink(*args):
    """
    Return the filename referenced by a sysfs symbolic link. If the
    symbolic link doesn't exist, this returns None.

    """
    linkname = _sysfs_filearg(args)
    if not os.path.exists(linkname):
        return None
    if not os.path.islink(linkname):
        raise Exception("%s is not a symlink" % linkname)
    data = os.readlink(linkname)
    return data


def sysfs_readlink_base(*args):
    """
    Return the basename of the file referenced by a sysfs symbolic
    link. If the symbolic link doesn't exist, this returns None.

    """
    linkname = _sysfs_filearg(args)
    data = sysfs_readlink(linkname)
    if data is None:
        return None
    else:
        return os.path.basename(data)


def sysfs_readstr_int(*args):
    """
    Read the content of a sysfs file as a string representing an
    integer value.

    """
    path = _sysfs_filearg(args)
    data = sysfs_read(path)
    if data is None:
        return None
    else:
        return int(data)


def sysfs_readstr_hex(*args):
    """
    Read the content of a sysfs file as a string representing an
    hexadecimal value.

    """
    path = _sysfs_filearg(args)
    data = sysfs_read(path)
    if data is None:
        return None
    else:
        return int(data, 16)


def sysfs_resolve_mdpath(mdpath):
    """
    Convert a PCI device path, as defined in a guest MD, to the path
    to the corresponding device under /sys.

    For example, /pci@300/pci@1/pci@0/pci@1/network@0 becomes
    /sys/bus/pci/devices/0001:03:00.0

    """
    return _SYSFS_MD_PATHMAP.get(mdpath)


def sysfs_walk_pci_devices(func, data=None):
    """
    Invoke the specified function for each PCI devices.

    """
    for dev in glob.glob(_SYSFS_PCIDEV_PATTERN):
        func(dev, data)


def sysfs_pcidev_get_vf(dev, vfid):
    """
    Return the PCI device corresponding to a SR-IOV VF of a PCI device.

    """
    vf_dev = sysfs_readlink_base(dev, "virtfn%d" % vfid)
    if vf_dev is None:
        return None
    else:
        return _SYSFS_PCI_DEVICES + "/" + vf_dev


def sysfs_pcidev_get_net(dev, default=None):
    """
    Return network interfaces of a PCI device.

    """
    net_path = glob.glob(dev + "/net/*")
    if not net_path:
        return default

    return list(os.path.basename(net) for net in net_path)


def sysfs_pcidev_get_driver(dev, default=None):
    """
    Return the driver name of a PCI device.

    """
    driver = sysfs_readlink_base(dev, "driver")
    if driver is None:
        return default

    return driver


def sysfs_pcidev_bind(dev, driver):
    """
    Bind a driver to a PCI device.

    """
    slot = os.path.basename(dev)
    sysfs_write("/sys/bus/pci/drivers/%s/bind" % driver, slot)


def sysfs_pcidev_unbind(dev):
    """
    Unbind any driver bound to a PCI device.

    """
    driver = sysfs_pcidev_get_driver(dev)
    if driver is None:
        return

    slot = os.path.basename(dev)
    sysfs_write("/sys/bus/pci/drivers/%s/unbind" % driver, slot)


def sysfs_pcidrv_newid(drv, vendor, device):
    """
    Add a new device ID to a PCI device driver.

    """
    sysfs_write("/sys/bus/pci/drivers/%s/new_id" % drv,
                "%x %x" % (vendor, device))


def sysfs_sriov_pfs():
    """
    Return the list of SR-IOV Physical Functions.

    """
    pf_list = glob.glob(_SYSFS_PCIDEV_SRIOV_NUMVFS_PATTERN)
    return list(os.path.dirname(pf) for pf in pf_list)


def sysfs_sriov_vfs(pf):
    """
    Return the list of SR-IOV Virtual Functions for a PF.

    """
    pattern = pf + "/virtfn*"
    vfid_pos = len(pattern) - 1
    vf_list = glob.glob(pattern)
    return list((int(vf[vfid_pos:]), sysfs_readlink_base(vf))
                for vf in vf_list)


_sysfs_init()
